import shutil
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Set, Iterator, Iterable

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distinfo import DistInfo, RecordsFileConfiguration
from pkm.api.distributions.pth_link import PthLink
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.packages.site_packages import InstalledPackage
from pkm.api.repositories.repository import AbstractRepository, Repository
from pkm.distributions.executables import Executables
from pkm.utils.commons import NoSuchElementException
from pkm.utils.files import CopyTransaction, is_empty_directory, is_relative_to
from pkm.utils.iterators import first_or_none


class SharedPackagesRepository(AbstractRepository):

    def __init__(self, workspace: Path, base_repository: Repository):
        super().__init__("shared")
        self._workspace = workspace
        self._base_repo = base_repository

    def remove_unused_packages(self, package_users: Iterator[Environment]):
        used_packages: Set[str] = set()
        for package_user in package_users:
            for used_package in package_user.site_packages.installed_packages():
                if (shared_marker := used_package.dist_info.path / "SHARED").exists():
                    used_packages.add(shared_marker.read_text())

        for package_dir in self._workspace.iterdir():
            if not package_dir.is_dir():
                continue

            for version_dir in package_dir.iterdir():
                if not version_dir.is_dir():
                    continue

                for shared_dir in version_dir.iterdir():
                    if not shared_dir.is_dir():
                        continue

                    if str(shared_dir.absolute()) not in used_packages:
                        shutil.rmtree(shared_dir)

                if is_empty_directory(version_dir):
                    version_dir.rmdir()

            if is_empty_directory(package_dir):
                package_dir.rmdir()

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        packages_dir = self._workspace / dependency.package_name
        packages = self._base_repo.match(dependency, env)
        return [
            _SharedPackage(p, packages_dir / str(p.version))
            for p in packages
        ]

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        def key(p: Package):
            return 0 if isinstance(p, _SharedPackage) else 1

        packages.sort(key=key)
        return packages

    def accepted_url_protocols(self) -> Iterable[str]:
        return self._base_repo.accepted_url_protocols()

    def accept_non_url_packages(self) -> bool:
        return self._base_repo.accept_non_url_packages()


@dataclass
class _SharedPackageArtifact:
    path: Path
    metadata: PackageMetadata
    compatibility_tags: str


class _SharedPackage(Package):

    def __init__(self, package: Package, shared_path: Path):
        self._package = package
        self._shared_path = shared_path

        if shared_path.exists():
            self._artifacts = [
                _SharedPackageArtifact(artifact, PackageMetadata.load(artifact / "dist-info/METADATA"), artifact.name)
                for artifact in shared_path.iterdir()
            ]
        else:
            self._artifacts = []

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._package.descriptor

    def _shared_artifact_for(self, env: Environment) -> Optional[_SharedPackageArtifact]:
        try:
            return max(
                ((a, s) for a in self._artifacts
                 if (s := env.compatibility_tag_score(a.compatibility_tag)) is not None),
                key=lambda it: it[1]
            )[0]
        except ValueError:
            return None

    def dependencies(
            self, environment: "Environment",
            extras: Optional[List[str]] = None) -> List["Dependency"]:

        if artifact := self._shared_artifact_for(environment):
            deps = artifact.metadata.dependencies
        else:
            deps = self._package.dependencies(environment)

        return [d for d in deps if d.is_applicable_for(environment, extras)]

    def is_compatible_with(self, env: "Environment") -> bool:
        return self._package.is_compatible_with(env)

    def install_to(
            self, target: "PackageInstallationTarget", user_request: Optional["Dependency"] = None,
            editable: bool = False):
        if shared := self._shared_artifact_for(target.env):
            _link_shared(self.descriptor, shared, target)
        else:
            self._package.install_to(target, user_request, editable)
            if shared := _move_to_shared(self._package.descriptor, target, self._shared_path):
                _link_shared(self.descriptor, shared, target)


def _copy_records(root: Path, shared: Path, records: List[Path]) -> List[Path]:
    shared.mkdir(parents=True)
    records_left = []
    for record in records:
        if is_relative_to(record, root):  # the path may lead out of the required source
            shared_path = shared / record.relative_to(root)
            if record.is_dir():
                shared_path.mkdir(exist_ok=True, parents=True)
            elif record.exists():
                shared_path.parent.mkdir(exist_ok=True, parents=True)
                shutil.copy(record, shared_path)
        else:
            records_left.append(record)
    return records_left


def _move_to_shared(package: PackageDescriptor,
                    target: PackageInstallationTarget, shared_path: Path) -> Optional[_SharedPackageArtifact]:
    psname = (package.expected_src_package_name + "-").lower()

    for site in ('purelib', 'platlib'):
        dist_info_path = first_or_none(
            it for it in Path(getattr(target, site)).glob("*.dist-info") if it.name.lower().startswith(psname))

        if dist_info_path:
            break
    else:
        raise NoSuchElementException(f"package: {package} is reported as installed but could not be find inside venv")

    dist_info = DistInfo.load(dist_info_path)

    iinfo = dist_info.load_installation_info()
    if iinfo.compatibility_tag:
        compatibility_tags = iinfo.compatibility_tag
        shared_target = shared_path / compatibility_tags
    else:
        return None  # TODO should probably notify the user that sharing is impossible for this package

    records = list(dist_info.installed_files())

    # filter dist info records:
    records = [
        r for r in records
        if not is_relative_to(r, dist_info.path)]

    # copy site
    shared_purelib = shared_target / site
    records = _copy_records(dist_info.path.parent, shared_purelib, records)

    # copy scripts
    script_entrypoints = {e.name for e in dist_info.load_entrypoints_cfg().entrypoints if e.is_script()}
    bin_path = Path(target.scripts)
    shared_bin = shared_target / "bin"

    # filter script entrypoints - they should be re-generated
    records = [r for r in records if not is_relative_to(r, bin_path) or r.stem not in script_entrypoints]
    records = _copy_records(bin_path, shared_bin, records)

    # copy dist-info
    shutil.copytree(dist_info.path, shared_target / 'dist-info')

    if records:
        warnings.warn(f"{len(records)} unsharable records found in package: {package.name} {package.version}: "
                      f"{', '.join(str(r) for r in records)}")

    # uninstall the original package
    InstalledPackage(dist_info).uninstall()

    return _SharedPackageArtifact(shared_target, dist_info.load_metadata_cfg(), shared_target.name)


def _link_shared(package: PackageDescriptor, shared: _SharedPackageArtifact, target: PackageInstallationTarget):
    with CopyTransaction() as ct:

        package_prefix = f"{package.expected_src_package_name}-{package.version}"

        site_name = "purelib" if (shared.path / "purelib").exists() else "platlib"
        site_path = Path(getattr(target, site_name))

        # first link the site data
        purelib_link = PthLink(site_path / f"{package_prefix}.pth", [shared.path / site_name])
        purelib_link.save()
        ct.touch(purelib_link.path)

        # now create all script entrypoints
        shared_distinfo = DistInfo.load(shared.path / "dist-info", non_standard_name_ok=True)
        bin_dir = Path(target.scripts)
        for entrypoint in shared_distinfo.load_entrypoints_cfg().entrypoints:
            if entrypoint.is_script():
                ct.touch(Executables.generate_for_entrypoint(target.env, entrypoint, bin_dir))

        # then, patch non entrypoint scripts
        for script in (shared.path / 'bin').iterdir():
            target_script = bin_dir / script.name
            Executables.patch_shabang_for_interpreter(script, target_script, target.env.interpreter_path)
            ct.touch(target_script)

        # we are almost done, copy dist-info
        distinfo_path = site_path / f"{package_prefix}.dist-info"
        ct.copy_tree(shared_distinfo.path, distinfo_path, lambda it: it.name != "RECORD")
        shared_marker = (distinfo_path / "SHARED")
        shared_marker.write_text(str(shared.path.absolute()))
        ct.touch(shared_marker)

        # and finally, sign the installation
        record_path = distinfo_path / "RECORD"
        RecordsFileConfiguration.load(record_path).sign_files(ct.copied_files, site_path).save()



from __future__ import annotations
import configparser
from abc import ABC, abstractmethod
from copy import copy, deepcopy
from pathlib import Path
from typing import Optional, Dict, Any, List, Sequence, Mapping, Iterator, Callable, TypeVar, Tuple, Type, cast, \
    MutableMapping, Generic, Iterable

from pkm.config import toml
from pkm.utils.commons import UnsupportedOperationException


class Configuration:

    def __init__(
            self, *,
            parent: Optional["Configuration"] = None,
            data: Optional[Dict[str, Any]] = None):
        self._parent = parent
        self._data: Dict[str, Any] = data if data is not None else {}

    def __getitem__(self, item: Sequence[str]) -> Any:
        if isinstance(item, str):
            item = toml.key2path(item)

        r = self._get(item)
        if r is None and self._parent:
            r = self._parent[item]
        return r

    def __contains__(self, item: Sequence[str]):
        return self[item] is not None

    def _get(self, path: Sequence[str]) -> Any:
        r = self._data
        for p in path:
            if r is None:
                return None
            if not isinstance(r, Mapping):
                raise ValueError(f"path: {path} passing through a terminal value {r} at '{p}'")

            r = r.get(p)

        return r

    def _subs_chain(self) -> Iterator["Configuration"]:
        result = []
        r = self
        while r is not None:
            result.append(r)
            r = r._parent

        return reversed(result)

    def collect(self, path: Sequence[str]) -> Any:
        if isinstance(path, str):
            path = toml.key2path(path)

        result = self[path]
        if isinstance(result, List):
            return [it for conf in self._subs_chain() for it in (conf._get(path) or [])]
        elif isinstance(result, Mapping):
            return {k: v for conf in self._subs_chain() for k, v in (conf._get(path).items() or {})}

        return result

    def with_parent(self, new_parent: Optional["Configuration"]) -> "Configuration":
        cp = copy(self)
        cp._parent = new_parent
        return cp

    def items(self) -> Iterable[Tuple[str, Any]]:
        return cast(Iterable[Tuple[str, Any]], self._data.items())


class MutableConfiguration(Configuration, ABC):
    def __setitem__(self, key: Sequence[str], value: Any):
        if isinstance(key, str):
            key = toml.key2path(key)

        self._set(key, value)

    def __delitem__(self, key: Sequence[str]):
        if isinstance(key, str):
            key = toml.key2path(key)

        self._del(key)

    @abstractmethod
    def save_to(self, path: Optional[Path] = None):
        ...

    def save(self) -> bool:
        self.save_to(None)
        return True

    def _set(self, path: Sequence[str], value: Any):
        r = self._data
        for p in path[:-1]:
            if not isinstance(r, MutableMapping):
                raise ValueError(f"path: {path} passing through a terminal value {r} at '{p}'")

            rp = r.get(p)
            if rp is None:
                r[p] = r = {}
            else:
                r = rp

        r[path[-1]] = value

    def get_or_put(self, key: Sequence[str], value_computer: Callable[[], Any]) -> Any:
        if isinstance(key, str):
            key = toml.key2path(key)

        result = self[key]
        if result is None:
            self[key] = result = value_computer()

        return result

    def _del(self, path: Sequence[str]):
        r = self._data
        for p in path[:-1]:
            if r is None:
                return

            if not isinstance(r, Mapping):
                raise ValueError(f"path: {path} passing through a terminal value {r} at '{p}'")

            r = r.get(p)

        if r is not None:
            del r[path[-1]]


_T = TypeVar('_T')
_C1R = TypeVar("_C1R", bound=Callable[[Any], Any])
_C1N = TypeVar("_C1N", bound=Callable[[Any], None])


class _ComputedConfigValue(Generic[_T]):

    def __init__(self, func: Callable[[Any], _T], dependency_keys: Tuple[str, ...]):
        self._func = func
        self._dependency_keys = tuple(toml.key2path(key) for key in dependency_keys)
        self.__doc__ = func.__doc__
        self._setter = None

    def __set_name__(self, owner: Type, name):
        if not issubclass(owner, Configuration):
            raise ValueError(
                "computed_config_value decorator can only be applied on methods of configuration derivatives")
        self._attr = f"__computed_{name}"
        self._stamp_attr = f"__computed_{name}_stamp"
        self._configuration = owner

    def modifier(self, func: _C1N) -> _C1N:
        self._setter = func
        return func

    def __set__(self, instance, value: _T):
        if instance is None or self._setter is None:
            raise UnsupportedOperationException('immutable field')

        self._setter(instance, value)
        setattr(instance, self._stamp_attr, None)

    def __get__(self, instance, owner) -> _T:
        if instance is None:
            return self

        new_stamp = tuple(id(instance[d]) for d in self._dependency_keys)
        try:
            old_stamp = getattr(instance, self._stamp_attr)
            if old_stamp == new_stamp:
                return getattr(instance, self._attr)
        except AttributeError:
            ...

        new_value = self._func(instance)
        setattr(instance, self._attr, new_value)
        setattr(instance, self._stamp_attr, new_stamp)
        return new_value


def computed_based_on(*based_on_keys: str) -> Callable[[Callable[[Any], _T]], _ComputedConfigValue[_T]]:
    def _computed(func: Callable[[Any], Any]):
        return _ComputedConfigValue(func, based_on_keys)

    return _computed


class FileConfiguration(MutableConfiguration, ABC):
    def __init__(self, *, path: Optional[Path], parent: Optional["Configuration"] = None,
                 data: Optional[Dict[str, Any]] = None):
        super().__init__(parent=parent, data=data)
        self._path = path

    def __deepcopy__(self, memodict={}):
        return type(self)(path=self.path, parent=deepcopy(self._parent), data=deepcopy(self._data))

    def exists_on_disk(self) -> bool:
        return self._path.exists()

    @abstractmethod
    def generate_content(self) -> str:
        ...

    def save_to(self, path: Optional[Path] = None):
        path = path or self._path
        path.parent.mkdir(exist_ok=True, parents=True)
        path.write_text(self.generate_content())

    def exists(self) -> bool:
        return self._path.exists()

    @property
    def path(self) -> Path:
        return self._path


class TomlFileConfiguration(FileConfiguration):

    def __init__(self, *, path: Optional[Path], parent: Optional["Configuration"] = None,
                 data: Optional[Dict[str, Any]] = None, dumps: Optional[Callable] = None):
        super().__init__(path=path, parent=parent, data=data)
        self._dumps = dumps

    def generate_content(self) -> str:
        dumps = self._dumps or toml.dumps
        return dumps(self._data)

    @classmethod
    def load(cls, file: Path, parent: Optional[Configuration] = None) -> "TomlFileConfiguration":
        data, dumps = toml.load(file) if file.exists() else ({}, None)
        return cls(path=file, parent=parent, data=data, dumps=dumps)


class InMemConfiguration(MutableConfiguration):

    def save_to(self, path: Optional[Path] = None):
        pass

    @classmethod
    def load(cls, data: Dict[str, Any], parent: Optional[Configuration]) -> "InMemConfiguration":
        return cls(parent=parent, data=data)


class _CaseSensitiveConfigParser(configparser.ConfigParser):
    optionxform = staticmethod(str)


class IniFileConfiguration(FileConfiguration):
    def generate_content(self) -> str:
        class StringWriter:
            def __init__(self):
                self.v: List[str] = []

            def write(self, s: str):
                self.v.append(s)

        sw = StringWriter()
        cp = _CaseSensitiveConfigParser()

        for key_or_section, value_or_content in self._data.items():
            if isinstance(value_or_content, Mapping):
                cp.add_section(key_or_section)
                for key, value in value_or_content.items():
                    cp.set(key_or_section, key, str(value))
            else:
                cp.set("", key_or_section, value_or_content)

        cp.write(sw)
        return ''.join(sw.v)

    @classmethod
    def load(cls, file: Path) -> "IniFileConfiguration":
        data = {}
        if file.exists():
            cp = _CaseSensitiveConfigParser()
            cp.read(str(file))
            for section_name, section_values in cp.items():
                if section_name and section_name != cp.default_section:
                    data[section_name] = {**section_values}
                else:
                    data.update(section_values)

        return cls(path=file, data=data)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict, Counter
from dataclasses import dataclass, replace
from itertools import chain
from typing import List, Dict, Iterable, Tuple, Optional, cast, DefaultDict, Union, Any, TypeVar, Protocol, Generic, Set

from pkm.api.versions.version import Version, StandardVersion
from pkm.api.versions.version_specifiers import VersionSpecifier, VersionMatch, AllowAllVersions, \
    RestrictAllVersions
from pkm.resolution.resolution_monitor import DependencyResolutionIterationEvent, \
    DependencyResolutionConclusionEvent, DependencyResolutionMonitoredOp
from pkm.utils.iterators import find_first
from pkm.utils.sequences import argmax


class PKG(Protocol):
    def __hash__(self): ...

    def __eq__(self, other): ...

    def __str__(self): ...

    def __lt__(self, other): ...


PKG_T = TypeVar('PKG_T', bound=PKG)


class UnsolvableProblemException(Exception):
    def __init__(self, incompatibility: "Incompatibility"):
        super().__init__(f'Dependency Versions Resolution Failed\n{incompatibility.report()}')
        self.incompatibility = incompatibility


class MalformedPackageException(Exception):
    ...


@dataclass(frozen=True)
class Term:
    package: PKG
    constraint: VersionSpecifier

    def negate(self) -> "Term":
        return Term(self.package, self.constraint.inverse())

    def intersect(self, other: "Term") -> "Term":
        assert self.package == other.package, 'cannot intersect terms of different packages'
        return Term(self.package, self.constraint.intersect_with(other.constraint))

    def satisfies(self, constraint: VersionSpecifier) -> bool:
        """self ⊆ term"""
        return constraint.allows_all(self.constraint)

    @classmethod
    def join(cls, package: PKG, terms: Iterable["Term"]) -> "Term":
        terms_iter = iter(terms)
        first_term = next(terms_iter, None)

        if not first_term:
            return Term.create(package, "*")

        next_term = next(terms_iter, None)
        if not next_term:
            return first_term

        package = first_term.package
        constraint = first_term.constraint

        while next_term is not None:
            constraint = constraint.intersect_with(next_term.constraint)
            next_term = next(terms_iter, None)

        return cls(package, constraint)

    def __repr__(self):
        return f"[{self.package} {self.constraint}]"

    @classmethod
    def create(cls, package: PKG, constraint: str) -> "Term":
        return cls(package, VersionSpecifier.parse(constraint))


@dataclass
class Assignment:
    term: Term
    decision_level: int
    order_index: int
    cause: Optional["Incompatibility"]
    accumulated: VersionSpecifier

    def is_decision(self) -> bool:
        return self.cause is None

    def __repr__(self):
        return f"A({self.term}, dlevel={self.decision_level}, cause={self.cause})"


class PartialSolution:
    def __init__(self, root_package: PKG = 'root'):
        self._root_package = root_package
        self._assignments_by_order: List[Assignment] = []
        self.assignments_by_package: DefaultDict[PKG, List[Assignment]] = defaultdict(list)
        self._required_packages: Dict[PKG, int] = {}
        self._decisions: Dict[PKG, Assignment] = {}

    def notify_state(self, current_package: PKG, monitor: DependencyResolutionMonitoredOp):
        # noinspection PyTypeChecker
        DependencyResolutionIterationEvent(self._decisions.keys(), self._required_packages.keys(), current_package) \
            .notify(monitor)

    def undecided_packages(self) -> List[PKG]:
        return [req for req in self._required_packages if req not in self._decisions]

    def requirement_decision_level(self, package: PKG) -> int:
        """
        returns the decision level in which the given `package` was required
        :param package: the package to check
        :return: the decision level if the package is required, otherwise -1
        """
        return self._required_packages.get(package) or -1

    @property
    def _decision_level(self) -> int:
        return max(0, len(self._decisions) - 1)

    def requiering_decision(self, package: PKG):
        decision_level = self._required_packages[package]
        return next(acc for acc in self._decisions.values()
                    if acc.decision_level == decision_level)

    def backtrack(self, decision_level: int):
        print(f"backtrack to decision_level: {decision_level}")

        def filtered(lst: List[Assignment]) -> Iterable[Assignment]:
            return (ass for ass in lst if ass.decision_level <= decision_level)

        self._assignments_by_order = list(filtered(self._assignments_by_order))

        assignments_by_package = defaultdict(list)
        for package, assignments in self.assignments_by_package.items():
            assignments_by_package[package].extend(filtered(assignments))

        self.assignments_by_package = assignments_by_package

        self._required_packages = {
            package: req_decision_level
            for package, req_decision_level in self._required_packages.items()
            if req_decision_level <= decision_level
        }

        self._decisions = {
            package: decision
            for package, decision in self._decisions.items()
            if decision.decision_level <= decision_level
        }

    def require(self, packages: Iterable[PKG]):
        required_packages = self._required_packages
        for package in packages:
            if package not in required_packages:
                required_packages[package] = self._decision_level

    def requires(self, package: PKG) -> bool:
        return package in self._required_packages

    def make_assignment(self, assignment_value: Term, cause: Optional["Incompatibility"] = None):
        package_assignments = self.assignments_by_package[assignment_value.package]

        accumulated_constraint = assignment_value.constraint
        if package_assignments:
            accumulated_constraint = package_assignments[-1].accumulated.intersect_with(accumulated_constraint)

        dlevel = self._decision_level

        if cause is None and assignment_value.package != self._root_package:
            dlevel += 1

        return Assignment(assignment_value, dlevel, len(self._assignments_by_order), cause, accumulated_constraint)

    def assign(self, assignment_value: Union[Term, Assignment], cause: Optional["Incompatibility"] = None):
        assignment = self.make_assignment(assignment_value, cause) \
            if not isinstance(assignment_value, Assignment) else assignment_value

        self.assignments_by_package[assignment.term.package].append(assignment)
        self._assignments_by_order.append(assignment)

        if assignment.is_decision():
            print(f"decided: {assignment}")
            self._decisions[assignment.term.package] = assignment
        else:
            print(f"derrived: {assignment}")
            ...

    def __repr__(self):
        return f"PartialSolution({self._assignments_by_order})"

    def decisions(self) -> Dict[PKG, Version]:
        return {package: cast(VersionMatch, ass.term.constraint).version for package, ass in self._decisions.items()}


@dataclass
class IncompatibilitySatisfaction:
    incompatibility: "Incompatibility"
    satisfier: Optional[Assignment] = None
    prev_satisfier: Optional[Assignment] = None
    undecided_term: Optional[Term] = None

    def is_full(self) -> bool:
        return bool(self.satisfier)

    def is_almost_full(self) -> bool:
        return bool(self.undecided_term)

    def __str__(self):
        if self.is_full():
            return f"fully-satisfied: sat={self.satisfier}, pre-sat={self.prev_satisfier}"
        if self.is_almost_full():
            return f"almost-fully-satisfied: undecided={self.undecided_term}"
        return "not-satisfied"

    @staticmethod
    def find_satisfier(
            solution: "PartialSolution", term: Term, next_satisfier: Optional[Assignment] = None
    ) -> Optional[Assignment]:

        if package_assignments := solution.assignments_by_package[term.package]:
            if term.constraint.allows_all(package_assignments[-1].accumulated):
                for assignment in package_assignments:
                    if assignment is next_satisfier:
                        return None

                    acc = assignment.accumulated
                    if next_satisfier:
                        if term.constraint.allows_all(acc.intersect_with(next_satisfier.term.constraint)):
                            return assignment
                    elif term.constraint.allows_all(acc):
                        return assignment
        return None

    @classmethod
    def create_unsatisfied(cls, incompatibility: Incompatibility) -> IncompatibilitySatisfaction:
        return cls(incompatibility)

    @classmethod
    def create_almost_full(cls, incompatibility: Incompatibility, undecided_term: Term) -> IncompatibilitySatisfaction:
        return cls(incompatibility, undecided_term=undecided_term)

    @classmethod
    def create_full(
            cls, incompatibility: Incompatibility, solution: PartialSolution, satisfiers: List[Assignment]
    ) -> IncompatibilitySatisfaction:
        satisfier_index = argmax(satisfiers, key=lambda s: s.order_index)
        satisfier = satisfiers.pop(satisfier_index)

        for term in incompatibility.terms:
            if solution.requires(term.package):
                satisfiers.append(solution.requiering_decision(term.package))

            if term.package == satisfier.term.package:
                if prev_satisfier := cls.find_satisfier(solution, term, satisfier):
                    satisfiers.append(prev_satisfier)

        prev_satisfier_index = argmax(satisfiers, key=lambda s: s.order_index) if satisfiers else None

        return cls(
            incompatibility, satisfier=satisfier,
            prev_satisfier=satisfiers[prev_satisfier_index] if prev_satisfier_index is not None else None)


class Incompatibility:
    def __init__(self, terms: Tuple[Term, ...], internal_cause: Optional[Tuple["Incompatibility", "Incompatibility"]],
                 external_cause: Optional[str]):
        assert not internal_cause or internal_cause[0] != internal_cause[1], "duplicated internal cause"

        self.terms = terms
        self.internal_cause = internal_cause
        self.external_cause = external_cause

    def is_simple(self):
        if self.internal_cause:
            ic1, ic2 = self.internal_cause
            return ic1.external_cause and ic2.external_cause

        return False

    def term_for(self, package: PKG) -> Optional[Term]:
        return next((term for term in self.terms if term.package == package), None)

    def check_satisfaction(self, solution: "PartialSolution") -> IncompatibilitySatisfaction:
        assignments = solution.assignments_by_package
        undecided_term: Optional[Term] = None
        satisfiers: List[Assignment] = []

        for term in self.terms:
            package_assignments = assignments[term.package]
            if package_assignments and package_assignments[-1].accumulated is RestrictAllVersions:
                return IncompatibilitySatisfaction.create_unsatisfied(self)

        for term in self.terms:
            satisfier = IncompatibilitySatisfaction.find_satisfier(solution, term)
            if not satisfier and (assignments := solution.assignments_by_package[term.package]) \
                    and not assignments[-1].accumulated.allows_any(term.constraint):
                return IncompatibilitySatisfaction.create_unsatisfied(self)

            if not satisfier:
                if not undecided_term:
                    undecided_term = term
                else:
                    return IncompatibilitySatisfaction.create_unsatisfied(self)
            else:
                satisfiers.append(satisfier)

        if undecided_term:
            return IncompatibilitySatisfaction.create_almost_full(self, undecided_term)

        return IncompatibilitySatisfaction.create_full(self, solution, satisfiers)

    def is_empty(self) -> bool:
        return not self.terms

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Incompatibility) and o.terms == self.terms

    def __hash__(self):
        return hash(self.terms)

    def __str__(self):
        if self.external_cause:
            return self.external_cause

        terms = self.terms
        if len(terms) == 1:
            if terms[0].constraint is AllowAllVersions:
                return f"{terms[0].package} cannot be resolved"
            return f"{terms[0].package} must be {terms[0].constraint.inverse()}"

        return f"{terms}"

    def __repr__(self):

        if self.external_cause:
            return f"EXI'{str(self)}'"
        else:
            return f"INI'{str(self)}'"

    def report(self):
        report: Dict[Any, Tuple[int, str]] = {}

        def write(key: Any, text: str):
            report[key] = len(report), text

        def generate(incompatability: Incompatibility):
            if incompatability.external_cause:
                write(incompatability, incompatability.external_cause)
                return

            ic1, ic2 = incompatability.internal_cause
            if ic1.internal_cause and ic2.internal_cause:
                l1, l2 = report.get(ic1), report.get(ic2)

                if l1 and l2:
                    write(incompatability, f"Because {ic1} ({l1[0]}) and {ic2} ({l2[0]}), {incompatability}")
                elif l1:
                    generate(ic2)
                    write(incompatability, f"And because {ic1} ({l1[0]}), {incompatability}")
                elif l2:
                    generate(ic1)
                    write(incompatability, f"And because {ic2} ({l2[1]}), {incompatability}")
                elif ic1.is_simple():
                    generate(ic2)
                    generate(ic1)
                    write(incompatability, f"Thus, {incompatability}")
                elif ic2.is_simple():
                    generate(ic1)
                    generate(ic2)
                    write(incompatability, f"Thus, {incompatability}")
                else:
                    generate(ic1)
                    write(len(report), '')
                    generate(ic2)
                    write(len(report), '')
                    write(incompatability,
                          f"So, because {ic1} ({report[ic1][0]}) and {ic2} ({report[ic2][0]}), {incompatability}")
            elif ic1.internal_cause or ic2.internal_cause:
                dr = ic1 if ic1.internal_cause else ic2
                ex = ic1 if ic2 is dr else ic2

                dl, el = report.get(dr), report.get(ex)
                if dl:
                    write(incompatability, f"Because {ex} and {dr} ({dl[0]}), {incompatability}")
                elif bool(dr.internal_cause[0].internal_cause) ^ bool(dr.internal_cause[1].internal_cause):
                    prior_dr = dr.internal_cause[0] if dr.internal_cause[0].internal_cause else dr.internal_cause[1]
                    prior_ex = dr.internal_cause[0] if prior_dr is dr.internal_cause[1] else dr.internal_cause[1]
                    generate(prior_dr)
                    write(incompatability, f"And because {prior_ex} and {ex}, {incompatability}")
                else:
                    generate(dr)
                    write(incompatability, f"And because {ex}, {incompatability}")
            else:
                if ic2.external_cause.startswith("Compatible version for"):
                    write(incompatability, f"Because {ic1} which is unresolved, {incompatability}")
                else:
                    write(incompatability, f"Because {ic1} and {ic2}, {incompatability}")

        generate(self)
        lines = sorted(report.values(), key=lambda x: x[0])
        return '\n'.join(f"{line}) {text}" for line, text in lines)

    @classmethod
    def create(cls, terms: Iterable[Term],
               internal_cause: Optional[Tuple["Incompatibility", "Incompatibility"]] = None,
               external_cause: Optional[str] = None) -> "Incompatibility":

        grouped_terms: Dict[PKG, List[Term]] = defaultdict(list)
        for term in terms:
            grouped_terms[term.package].append(term)

        normalized_terms = [Term.join(package, terms)
                            for package, terms in grouped_terms.items()]

        sorted_terms = sorted(normalized_terms, key=lambda x: str(x.package))
        return cls(tuple(sorted_terms), internal_cause, external_cause)

    def update_inavailability(self, new_constraint: VersionSpecifier):
        assert len(self.terms) == 1, 'attempting to update terms for non inavailability incompatability'
        self.terms = (replace(self.terms[0], constraint=new_constraint),)
        self.external_cause = f'Compatible version for {self.terms[0]} not found'


class DependencyIncompatibility(Incompatibility):
    def __init__(self, dependant: Term, dependency: Term):
        super().__init__(
            tuple(sorted([dependency.negate(), dependant], key=lambda x: str(x.package))), None,
            f"{dependant} depends on {dependency}")

        self.dependant = dependant
        self.dependency = dependency
        self.added = False

    def check_satisfaction(self, solution: "PartialSolution") -> IncompatibilitySatisfaction:
        if dependant_assignments := solution.assignments_by_package[self.dependant.package]:
            dependant_assignment = dependant_assignments[-1]
            acc = dependant_assignment.accumulated

            if acc.allows_any(self.dependant.constraint) or \
                    (acc is not RestrictAllVersions and self.dependant.constraint.allows_all(acc)):
                return super().check_satisfaction(solution)

        return IncompatibilitySatisfaction.create_unsatisfied(self)

    def update_dependant(self, new_dependant: Term):
        package = str(new_dependant.package)
        assert package == str(self.dependant.package), \
            "attempting to update dependency incompatibility with a term that does not belongs to it"

        self.dependant = new_dependant
        self.terms = tuple(sorted([self.dependency.negate(), new_dependant], key=lambda x: str(x.package)))
        self.external_cause = f"{new_dependant} depends on {self.dependency}"


@dataclass
class _PackageVersion:
    term: Term
    # the effective range of this version (v2.0.0 -> range >=1.7.3, <1.0.1)
    effective_constraint_range: VersionSpecifier = AllowAllVersions
    dependencies: Optional[Dict[PKG, "PackageDependency"]] = None
    next: Optional["_PackageVersion"] = None

    @property
    def version(self) -> Version:
        return cast(VersionMatch, self.term.constraint).version

    def fetch_dependencies_info(self, problem: Problem):
        if not self.dependencies:
            version_dependencies = problem.get_dependencies(self.term.package, self.version)
            self.dependencies = {
                d.package: PackageDependency(d)
                for d in version_dependencies}

    def compute_incompatibilities(self) -> List[Incompatibility]:
        result: List[Incompatibility] = []
        for dependency in self.dependencies.values():
            last_requiring = self

            while last_requiring.next:
                nxt = last_requiring.next
                nxt_deps = nxt.dependencies
                if not nxt_deps:
                    break

                nxt_dep = nxt_deps.get(dependency.term.package)
                if not nxt_dep or nxt_dep.term.constraint != dependency.term.constraint:
                    break

                last_requiring = nxt

            if last_requiring is not self:
                incompatibility = last_requiring.dependencies[dependency.term.package].incompatibility
                if incompatibility:
                    result.append(incompatibility)
                    incompatibility.update_dependant(
                        Term(self.term.package,
                             incompatibility.dependant.constraint.union_with(self.effective_constraint_range)))
                    continue

            nt = Term(self.term.package, self.effective_constraint_range)
            dependency.incompatibility = DependencyIncompatibility(nt, dependency.term)
            result.append(dependency.incompatibility)

        return result

    def __repr__(self):
        return f"V({self.term})"


@dataclass
class PackageDependency:
    term: Term
    incompatibility: Optional[DependencyIncompatibility] = None


class Problem(ABC):
    """
    represents the input for the pubgrub algorithm = the problem it attempts to solve.
    """

    @abstractmethod
    def get_dependencies(self, package: PKG, version: Version) -> List[Term]:
        """
        :param package: the package name to get dependencies to
        :param version: the required version of the `package`
        :return: list of dependencies for the given `package` and `version`
        """

    @abstractmethod
    def get_versions(self, package: PKG) -> List[StandardVersion]:
        """
        get the available (standard) versions of the given package in importance order
         (first is more important than last)
        :param package: the package name to look for version in
        :return: versions of the package ordered by importance
        """

    @abstractmethod
    def has_version(self, package: PKG, version: Version) -> bool:
        """
        checks for the existence of the given `version`
        :param package: the package to check
        :param version: the version to check
        :return: True if the given package and version pair could be found, False otherwise
        """


class Solver(Generic[PKG_T]):

    def __init__(self, problem: Problem, root_package: PKG_T = 'root'):
        self._root_package = root_package
        self._problem = problem
        self._solution = PartialSolution(root_package)
        self._package_versions: Dict[PKG, List[_PackageVersion]] = {}
        self._package_availability_incompatibilities: Dict[PKG, Incompatibility] = {}
        self._package_trouble_level: Counter[PKG] = Counter()

        # incompatibilities by package
        self._incompatibilities: DefaultDict[PKG, List[Incompatibility]] = defaultdict(list)

    def _add_incompatibility(self, incompatibility: Incompatibility):
        if isinstance(incompatibility, DependencyIncompatibility):
            if incompatibility.added:
                return True
            incompatibility.added = True

        print(f'adding incompatibility: {incompatibility}')
        for term in incompatibility.terms:
            term_incompatibilities = self._incompatibilities[term.package]
            if incompatibility not in term_incompatibilities:
                term_incompatibilities.append(incompatibility)

    def solve(self) -> Dict[PKG_T, Version]:
        with DependencyResolutionMonitoredOp() as mop:
            root_term = self.package_versions(self._root_package)[0].term
            self._add_incompatibility(
                Incompatibility.create([root_term.negate()], external_cause='Root Project'))

            self._solution.require([root_term.package])

            next_package = root_term.package
            while next_package is not None:
                self._solution.notify_state(next_package, mop)
                print(f"trying to solve for {next_package}, already decided on: {self._solution.decisions()}")
                self._propagate(next_package)
                next_package = self._make_next_decision()

            print(f"reached into conclusion: {self._solution.decisions()}")
            DependencyResolutionConclusionEvent(self._solution.decisions()).notify(mop)
            return self._solution.decisions()

    def _propagate(self, next_package: PKG):

        print("#### unit propagation ####")
        changed = {next_package}

        while changed:
            package = changed.pop()

            for incompatibility in list(reversed(self._incompatibilities[package])):
                satisfaction: IncompatibilitySatisfaction = incompatibility.check_satisfaction(self._solution)
                if satisfaction.is_full():
                    print(f"incompatibility: {incompatibility} satisfied, entering conflict resolution")
                    changed.add(self._resolve_conflict(incompatibility, satisfaction))

                elif satisfaction.is_almost_full():
                    term = satisfaction.undecided_term
                    print(f"incompatibility {incompatibility} is almost full, undecided_term is {term}")
                    self._solution.assign(term.negate(), incompatibility)
                    changed.add(term.package)

    def _is_tautology(self, incompatibility: Incompatibility) -> bool:
        return not (terms := incompatibility.terms) or \
               (len(terms) == 1 and terms[0].package == self._root_package)

    def _resolve_conflict(
            self, incompatibility: Incompatibility, satisfaction: IncompatibilitySatisfaction) -> PKG:

        print("#### conflict resolution ####")
        original_incompatibility = incompatibility

        while True:
            print(f"enter conflict resolution loop with {incompatibility}")
            for term in incompatibility.terms:
                self._package_trouble_level[term.package] += 1

            if self._is_tautology(incompatibility):
                raise UnsolvableProblemException(incompatibility)

            satisfier = satisfaction.satisfier
            term = incompatibility.term_for(satisfier.term.package)

            prev_satisfier = satisfaction.prev_satisfier
            prev_satisfier_level = prev_satisfier.decision_level if prev_satisfier else 0
            print(f"satisfier: {satisfier}, prev_satisfier: {prev_satisfier}")

            if satisfier.is_decision() or prev_satisfier_level < satisfier.decision_level:
                self._solution.backtrack(prev_satisfier_level)
                if incompatibility is not original_incompatibility:
                    self._add_incompatibility(incompatibility)

                satisfaction = incompatibility.check_satisfaction(self._solution)
                assert satisfaction.is_almost_full(), \
                    "new incompatibility resulted after conflict resolution is not almost satisfied"

                term = satisfaction.undecided_term
                print(f"conflict happened because {term}, rejecting this assignment from now on")
                self._solution.assign(term.negate(), incompatibility)
                return term.package

            prior_cause_terms: Set[Term] = {
                term for term in chain(incompatibility.terms, satisfier.cause.terms)
                if term.package != satisfier.term.package}

            assert satisfier.cause != incompatibility  # this massup the error reporting...
            if not satisfier.term.satisfies(term.constraint) or not prior_cause_terms:
                prior_cause_terms.add(
                    Term(satisfier.term.package,
                         # satisfier.term.constraint.difference_from(term.constraint).inverse()))
                         satisfier.accumulated.difference_from(term.constraint).inverse()))

            prev_incompatibility = incompatibility
            incompatibility = Incompatibility.create(
                list(prior_cause_terms), internal_cause=(satisfier.cause, incompatibility))

            if len(prior_cause_terms) == 0:
                self._solution.backtrack(0)
                self._solution.assign(Term(term.package, RestrictAllVersions), cause=incompatibility)
                return term.package

            print(f"root cause: {incompatibility}")

            if not self._is_tautology(incompatibility):
                satisfaction = incompatibility.check_satisfaction(self._solution)
                assert satisfaction.is_full(), \
                    f"incompatibility: {prev_incompatibility} resulted in " \
                    f"root cause: {incompatibility} that is not satisfied"

    def package_versions(self, package: PKG) -> List[_PackageVersion]:
        versions = self._package_versions.get(package)
        if not versions:
            versions = self._package_versions[package] = [
                _PackageVersion(Term(package, VersionMatch(ver)))
                for ver in self._problem.get_versions(package)]

            sorted_versions = sorted(versions, key=lambda v: v.version)

            for i in range(len(versions) - 1):
                sorted_versions[i].next = sorted_versions[i + 1]

            if len(sorted_versions) == 1:
                sorted_versions[0].effective_constraint_range = AllowAllVersions
            elif sorted_versions:
                for i in range(1, len(versions) - 1):
                    sorted_versions[i].effective_constraint_range = VersionSpecifier.create_range(
                        min_=cast(StandardVersion, sorted_versions[i - 1].version),
                        max_=cast(StandardVersion, sorted_versions[i].version), includes_max=True)

                sorted_versions[0].effective_constraint_range = VersionSpecifier.create_range(
                    max_=cast(StandardVersion, sorted_versions[0].version), includes_max=True)
                sorted_versions[-1].effective_constraint_range = VersionSpecifier.create_range(
                    min_=cast(StandardVersion, sorted_versions[-2].version))

        return versions

    def _attempt_minor_adjustments(self, package: PKG, conflicts: List[Incompatibility]) -> bool:
        # attempt the minor adjustment heuristic: the idea is that if this package cannot be selected
        # because we previously chose incompatible version of one of its dependencies then we check
        # what was the actual limitation of this dependency when we chose its version, if the limitation allows
        # the required version, we backtrack to remove this assignment and instead assign this package version
        # which will cause the dependency to be adjusted and chosen correctly

        minor_adjustment_dlevel = -1
        for conflict in conflicts:
            dependency: Term = find_first(conflict.terms, lambda it: it.package != package)
            dependency_assignments = self._solution.assignments_by_package[dependency.package]
            dependency_requirement = dependency.constraint.inverse()
            if len(dependency_assignments) > 1 and dependency_assignments[-1].is_decision():
                actual_limitation = dependency_assignments[-2].accumulated
                if actual_limitation.allows_any(dependency_requirement):
                    print(f"{package} is conflicting with {dependency.package}, it requires that ", end='')
                    print(f"{dependency_requirement} but it actual limitation in solution is '{actual_limitation}'")
                    dlevel = dependency_assignments[-1].decision_level - 1
                    minor_adjustment_dlevel = dlevel if minor_adjustment_dlevel == -1 \
                        else min(dlevel, minor_adjustment_dlevel)
                else:
                    minor_adjustment_dlevel = -1
                    break

        if minor_adjustment_dlevel >= 0 \
                and self._solution.requirement_decision_level(package) <= minor_adjustment_dlevel:
            print(f"applying minor adjusments heuristic - backtracking to {minor_adjustment_dlevel}")
            self._solution.backtrack(minor_adjustment_dlevel)
            return True

        return False

    def _make_next_decision(self) -> Optional[PKG]:
        print("#### decision ####")
        undecided_packages = self._solution.undecided_packages()
        if not undecided_packages:
            return None

        print(f"undecided packages: {undecided_packages}")
        package_matching_versions: Dict[PKG, List[_PackageVersion]] = {}
        for package in undecided_packages:
            acc_assignment = self._solution.assignments_by_package[package][-1].accumulated

            if isinstance(acc_assignment, VersionMatch) and acc_assignment.allow and \
                    not isinstance(acc_assignment.version, StandardVersion):

                if self._problem.has_version(package, acc_assignment.version):
                    package_matching_versions[package] = [
                        _PackageVersion(Term(package, acc_assignment), acc_assignment)]
                else:
                    package_matching_versions[package] = []
            else:
                versions = self.package_versions(package)
                package_matching_versions[package] = [pver for pver in versions if
                                                      acc_assignment.allows_version(pver.version)]

        package = min(undecided_packages,
                      key=lambda pack: (-self._package_trouble_level[pack], len(package_matching_versions[pack])))

        print(f"choosing to try and assign {package} with constraint: ", end='')
        print(f"{self._solution.assignments_by_package[package][-1].accumulated}")
        versions = list(package_matching_versions[package])  # defensive copy because we might change it

        while True:
            if not versions:
                acc_assignment = self._solution.assignments_by_package[package][-1].accumulated

                if ic := self._package_availability_incompatibilities.get(package):
                    ic.update_inavailability(ic.terms[0].constraint.union_with(acc_assignment))
                else:
                    print(f"could not find version that match {acc_assignment}")
                    ic = Incompatibility.create(
                        [Term(package, acc_assignment)],
                        external_cause=f'Compatible version for [{package} {acc_assignment}] not found')
                    self._package_availability_incompatibilities[package] = ic
                    self._add_incompatibility(ic)
                return package

            version = versions.pop(0)
            print(f"version: {version} match our term")

            try:
                version.fetch_dependencies_info(self._problem)
            except MalformedPackageException:
                import traceback
                traceback.print_exc()
                print(f"version: {version} discovered to be malformed")
                continue  # retry with another version
            break

        incompatibilities = self._add_dependency_incompatibilities(version)

        assignment: Assignment = self._solution.make_assignment(Term(package, VersionMatch(version.version)), None)
        assignments = self._solution.assignments_by_package
        assignments[package].append(assignment)
        print(f"checking if we can still assign {version} after the new incompatibilities: {incompatibilities}")
        conflicts = list(filter(lambda it: it.check_satisfaction(self._solution).is_full(), incompatibilities))
        assignments[package].pop()

        if conflicts and self._attempt_minor_adjustments(package, conflicts):
            conflicts = []

        if not conflicts:
            print("we can!")
            self._solution.assign(assignment)
            self._solution.require(version.dependencies.keys())
        else:
            print(f"we cant.. ({conflicts})")
            ...

        return package

    def _add_dependency_incompatibilities(self, version: _PackageVersion) -> List[Incompatibility]:
        incompatibilities = version.compute_incompatibilities()
        for incompatibility in incompatibilities:
            self._add_incompatibility(incompatibility)

        return incompatibilities



from __future__ import annotations

from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import List, Optional, Set

from pkm.api.versions.version import Version, StandardVersion, UrlVersion
from pkm.utils.commons import UnsupportedOperationException
from pkm.utils.hashes import HashBuilder
from pkm.utils.iterators import first_or_raise
from pkm.utils.seqs import seq


class VersionSpecifier(ABC):

    def allows_all(self, other: VersionSpecifier) -> bool:
        return self.intersect_with(other) == other

    def allows_any(self, other: VersionSpecifier) -> bool:
        return self.intersect_with(other) is not RestrictAllVersions

    def allows_pre_or_dev_releases(self) -> bool:
        return False

    @abstractmethod
    def allows_version(self, version: Version) -> bool:
        ...

    @abstractmethod
    def union_with(self, other: VersionSpecifier) -> VersionSpecifier:
        ...

    @abstractmethod
    def intersect_with(self, other: VersionSpecifier) -> VersionSpecifier:
        ...

    @abstractmethod
    def inverse(self) -> VersionSpecifier:
        ...

    def difference_from(self, other: VersionSpecifier) -> VersionSpecifier:
        return self.intersect_with(other.inverse())

    @abstractmethod
    def __str__(self) -> str:
        ...

    @abstractmethod
    def __lt__(self, other: VersionSpecifier) -> VersionSpecifier:
        ...

    @abstractmethod
    def __eq__(self, other: VersionSpecifier) -> bool:
        ...

    def __le__(self, other: VersionSpecifier):
        return self == other or self < other

    def __gt__(self, other: VersionSpecifier):
        return other < self

    def __ge__(self, other):
        return other <= self

    @classmethod
    def parse(cls, txt: str) -> VersionSpecifier:
        from pkm.api.versions.version_parser import parse_specifier
        return parse_specifier(txt)

    @classmethod
    def create_range(
            cls, min_: Optional[StandardVersion] = None, max_: Optional[StandardVersion] = None,
            includes_min: bool = False, includes_max: bool = False) -> VersionSpecifier:

        eq = min_ == max_
        assert not (eq and includes_min != includes_max), \
            f"{min_}(min) == {max_}(max) but includes_min != includes_max "

        if min_ is None and max_ is None:
            return AllowAllVersions

        if eq:
            if not includes_min or not includes_max:
                return RestrictAllVersions

            return VersionMatch(min_, includes_min)

        return StandardVersionRange(min_, max_, includes_min, includes_max)


@dataclass(unsafe_hash=True)
class _RestrictAll(VersionSpecifier):

    def inverse(self) -> VersionSpecifier:
        return AllowAllVersions

    def __lt__(self, other):
        return False

    def __eq__(self, other):
        return other is RestrictAllVersions

    def allows_version(self, other: Version) -> bool:
        return False

    def allows_any(self, other: VersionSpecifier) -> bool:
        return False

    def allows_all(self, other: VersionSpecifier) -> bool:
        return False

    def __str__(self):
        return "<none>"

    def union_with(self, other: VersionSpecifier) -> VersionSpecifier:
        return other

    def intersect_with(self, other: VersionSpecifier) -> VersionSpecifier:
        return self


RestrictAllVersions = _RestrictAll()


@dataclass(unsafe_hash=True)
class _AllowAll(VersionSpecifier):

    def inverse(self) -> VersionSpecifier:
        return RestrictAllVersions

    def __lt__(self, other):
        return other is not AllowAllVersions

    def __eq__(self, other):
        return other is AllowAllVersions

    def allows_all(self, other: VersionSpecifier) -> bool:
        if other is RestrictAllVersions:
            return other.allows_all(self)

        return True

    def allows_any(self, other: VersionSpecifier) -> bool:
        if other is RestrictAllVersions:
            return other.allows_any(self)

        return True

    def allows_version(self, other: Version) -> bool:
        return True

    def __str__(self):
        return "*"

    def union_with(self, other: VersionSpecifier) -> VersionSpecifier:
        return self

    def intersect_with(self, other: VersionSpecifier) -> VersionSpecifier:
        return other


AllowAllVersions = _AllowAll()


@dataclass
class VersionsUnion(VersionSpecifier):
    segments: List[VersionSpecifier]

    def __post_init__(self):
        assert len(self.segments) > 1
        assert all(it is not RestrictAllVersions and it is not AllowAllVersions for it in self.segments)
        self.segments.sort()

    def __hash__(self):
        return HashBuilder().ordered_seq(self.segments).build()

    def allows_pre_or_dev_releases(self) -> bool:
        return any(it.allows_pre_or_dev_releases() for it in self.segments)

    def inverse(self) -> VersionSpecifier:
        result = AllowAllVersions
        for segment in self.segments:
            result = result.intersect_with(segment.inverse())
        return result

    def allows_version(self, version: Version) -> bool:
        return any(s.allows_version(version) for s in self.segments)

    def __lt__(self, other: VersionSpecifier):
        return self.segments[0] < other

    def __eq__(self, other: VersionSpecifier):
        return isinstance(other, VersionsUnion) and self.segments == other.segments

    def __str__(self):

        segs = self.segments
        if len(segs) == 2:
            s1, s2 = segs
            if isinstance(s1, StandardVersionRange) and isinstance(s2, StandardVersionRange):
                if not s1.min and not s2.max and s1.max == s2.min:
                    return f"!={s1.max}"
        return seq(segs).str_join('; ')

    def union_with(self, other: VersionSpecifier) -> VersionSpecifier:
        if isinstance(other, (_AllowAll, _RestrictAll)):
            return other.union_with(self)

        new_segments = [*self.segments]
        other_segments = other.segments if isinstance(other, VersionsUnion) else [other]

        while other_segments:
            os = other_segments.pop()
            for i, es in enumerate(new_segments):
                us = os.union_with(es)
                if us is AllowAllVersions:
                    return AllowAllVersions
                if us is RestrictAllVersions:
                    break
                if not isinstance(us, VersionsUnion):
                    other_segments.append(us)
                    del new_segments[i]
                    break
            else:
                new_segments.append(os)

        return VersionsUnion(new_segments)

    def intersect_with(self, other: VersionSpecifier) -> VersionSpecifier:
        if isinstance(other, (_AllowAll, _RestrictAll)):
            return other.intersect_with(self)

        result = RestrictAllVersions
        for segment in self.segments:
            result = result.union_with(segment.intersect_with(other))
        return result


@dataclass
class HetroVersionIntersection(VersionSpecifier):
    blacklist: Set[Version]
    standard_spec: VersionSpecifier

    def __post_init__(self):
        assert len(self.blacklist) > 0, "blacklist must be larger than 0"
        assert self.standard_spec is not RestrictAllVersions

    def allows_pre_or_dev_releases(self) -> bool:
        return self.standard_spec.allows_pre_or_dev_releases() \
               or any(it.is_pre_or_dev_release() for it in self.blacklist)

    def __hash__(self):
        return HashBuilder().unordered_seq(self.blacklist).regular(self.standard_spec).build()

    def allows_version(self, version: Version) -> bool:
        if version in self.blacklist:
            return False

        return not isinstance(version, StandardVersion) or self.standard_spec.allows_version(version)

    def union_with(self, other: VersionSpecifier) -> VersionSpecifier:
        if isinstance(other, (_RestrictAll, _AllowAll, VersionsUnion)):
            return other.union_with(self)

        if isinstance(other, HetroVersionIntersection):
            standard_spec = self.standard_spec.union_with(other.standard_spec)
            if len(other.blacklist.union(self.blacklist)) != len(self.blacklist):
                return standard_spec

            return HetroVersionIntersection.create(
                self.blacklist if len(self.blacklist) < len(other.blacklist) else other.blacklist, standard_spec)

        if isinstance(other, VersionMatch):
            if isinstance(other.version, StandardVersion) and other.allow:
                return HetroVersionIntersection.create(self.blacklist, self.standard_spec.union_with(other))

            blacklist = set(self.blacklist)
            if other.allow:
                blacklist.remove(other.version)
            else:
                blacklist.add(other.version)

            return HetroVersionIntersection.create(blacklist, self.standard_spec)

        if isinstance(other, StandardVersionRange):
            return HetroVersionIntersection.create(self.blacklist, self.standard_spec.union_with(other))

        raise UnsupportedOperationException()

    def intersect_with(self, other: VersionSpecifier) -> VersionSpecifier:
        if isinstance(other, (_RestrictAll, _AllowAll, VersionsUnion)):
            return other.intersect_with(self)

        if isinstance(other, HetroVersionIntersection):
            return HetroVersionIntersection.create(  # yes - union the blacklists
                self.blacklist.union(other.blacklist), self.standard_spec.intersect_with(other.standard_spec))

        if isinstance(other, VersionMatch):
            if isinstance(other.version, StandardVersion) and other.allow:
                if self.standard_spec.allows_version(other.version) and other.version not in self.blacklist:
                    return other
                return RestrictAllVersions

            if other.allow:
                if other.version in self.blacklist:
                    return RestrictAllVersions
                return other

            return HetroVersionIntersection.create({*self.blacklist, other.version}, self.standard_spec)

        if isinstance(other, StandardVersionRange):
            return HetroVersionIntersection.create(self.blacklist, self.standard_spec.intersect_with(other))

        raise UnsupportedOperationException()

    def inverse(self) -> VersionSpecifier:
        union = [*(VersionMatch(it) for it in self.blacklist)]
        if self.standard_spec is not AllowAllVersions:
            union.append(self.standard_spec.inverse())
        return VersionsUnion(union)

    def __str__(self) -> str:
        result = ", ".join(f"!={it}" for it in self.blacklist)
        if self.standard_spec is not AllowAllVersions:
            result += f", {self.standard_spec}"
        return result

    def __lt__(self, other: VersionSpecifier) -> bool:
        if isinstance(other, (_RestrictAll, _AllowAll, VersionsUnion)):
            return not self >= other

        if isinstance(other, HetroVersionIntersection):
            if self.blacklist == other.blacklist:
                return self.standard_spec < other.standard_spec
            return sorted(self.blacklist) < sorted(other.blacklist)

        if isinstance(other, VersionMatch):
            if isinstance(other.version, StandardVersion):
                return self.standard_spec < other
            return other.allow or sorted(self.blacklist) < [other.version]

        if isinstance(other, StandardVersionRange):
            return self.standard_spec < other

        raise UnsupportedOperationException()

    def __eq__(self, other: VersionSpecifier) -> bool:
        return isinstance(other, HetroVersionIntersection) \
               and self.blacklist == other.blacklist and self.standard_spec == other.standard_spec

    @classmethod
    def create(cls, blacklist: Set[Version], standard_spec: VersionSpecifier) -> VersionSpecifier:
        blacklist = {it for it in blacklist if not isinstance(it, StandardVersion) or standard_spec.allows_version(it)}

        if standard_spec is RestrictAllVersions:
            return RestrictAllVersions
        if len(blacklist) == 0:
            return standard_spec
        if len(blacklist) == 1 and standard_spec is AllowAllVersions:
            return VersionMatch(first_or_raise(blacklist), False)

        return HetroVersionIntersection(blacklist, standard_spec)


@dataclass(unsafe_hash=True)
class VersionMatch(VersionSpecifier):
    version: Version
    allow: bool = True

    def allows_pre_or_dev_releases(self) -> bool:
        return self.allow and self.version.is_pre_or_dev_release()

    def inverse(self) -> VersionSpecifier:
        return VersionMatch(self.version, not self.allow)

    def __lt__(self, other: VersionSpecifier):
        if isinstance(other, (_RestrictAll, _AllowAll, VersionsUnion, HetroVersionIntersection)):
            return not self >= other

        if isinstance(other, VersionMatch):
            if self == other:
                return False
            if self.allow:
                return not other.allow
            return other.allow and self.version < other.version

        if isinstance(other, StandardVersionRange):
            if isinstance(self.version, StandardVersion):
                if self.allow:
                    return other.min and self.version < other.min
                return bool(other.min)

            return True  # uid versions comes before any standard ones

        raise NotImplemented()

    def __eq__(self, other: VersionSpecifier) -> bool:
        return isinstance(other, VersionMatch) \
               and self.allow == other.allow and self.version == other.version

    def allows_version(self, version: Version) -> bool:
        if not self.version.is_local():
            version = version.without_local()

        if self.allow:
            return self.version == version
        return self.version != version

    def __str__(self):
        v, a = self.version, self.allow

        if isinstance(v, UrlVersion):
            return f'@{v}' if a else f'!@{v}'
        elif isinstance(v, StandardVersion):
            return f'=={v}' if a else f'!={v}'
        return f'==={v}' if a else f'!=={v}'

    def union_with(self, other: VersionSpecifier) -> VersionSpecifier:
        if isinstance(other, (_RestrictAll, _AllowAll, HetroVersionIntersection, VersionsUnion)):
            return other.union_with(self)

        if isinstance(other, VersionMatch):
            if self == other:
                return self

            if self.version == other.version:
                if self.allow != other.allow:
                    return AllowAllVersions

        if isinstance(other, StandardVersionRange):
            if other.allows_version(self.version):
                return other if self.allow else AllowAllVersions

            if not self.allow:
                return self

            if not other.includes_min and self.version == other.min:
                return StandardVersionRange(other.min, other.max, True, other.includes_max)
            if not other.includes_max and self.version == other.max:
                return StandardVersionRange(other.min, other.max, other.includes_min, True)

        return VersionsUnion([self, other])

    def intersect_with(self, other: VersionSpecifier) -> VersionSpecifier:
        if isinstance(other, (_RestrictAll, _AllowAll, HetroVersionIntersection, VersionsUnion)):
            return other.intersect_with(self)

        if isinstance(other, VersionMatch):
            if self == other:
                return self

            my_ver, ot_ver = self.version, other.version
            if my_ver.is_local() != ot_ver.is_local():
                my_ver, ot_ver = my_ver.without_local(), ot_ver.without_local()

            if self.allow and other.allow:
                if my_ver == ot_ver:
                    return self if self.version.is_local() else other
                return RestrictAllVersions
            if self.allow or other.allow:
                if my_ver == ot_ver:
                    return RestrictAllVersions
                return self if self.allow else other
            return HetroVersionIntersection({self.version, other.version}, AllowAllVersions)

        if isinstance(other, StandardVersionRange) and isinstance(self.version, StandardVersion):
            if other.allows_version(self.version):
                if self.allow:
                    return self
                return HetroVersionIntersection({self.version}, other)

        return RestrictAllVersions if self.allow else other


@dataclass(unsafe_hash=True)
class StandardVersionRange(VersionSpecifier):
    min: Optional[StandardVersion]
    max: Optional[StandardVersion]
    includes_min: bool
    includes_max: bool

    def __post_init__(self):
        assert self.min or self.max, "either min or max should be provided"
        assert not self.includes_min or self.min, "cannot include min if min is not provided"
        assert not self.includes_max or self.max, "cannot include max if max is not provided"
        assert not self.min or not self.max or self.min < self.max, f"min ({self.min}) must be < max ({self.max}) "

    def allows_pre_or_dev_releases(self) -> bool:
        return (self.min and self.min.is_pre_or_dev_release()) or (self.max and self.max.is_pre_or_dev_release())

    def inverse(self) -> VersionSpecifier:
        parts = []
        if self.min:
            parts.append(VersionSpecifier.create_range(None, self.min, False, self.min and not self.includes_min))
        if self.max:
            parts.append(VersionSpecifier.create_range(self.max, None, self.max and not self.includes_max, False))

        if len(parts) == 1:
            return parts[0]
        return parts[0].union_with(parts[1])

    def __lt__(self, other):
        if isinstance(other, (_RestrictAll, _AllowAll, VersionsUnion, HetroVersionIntersection, VersionMatch)):
            return not self >= other

        if isinstance(other, StandardVersionRange):
            if self == other:
                return False

            if self.min and other.min:
                if self.min != other.min:
                    return self.min < other.min
                elif self.includes_min != other.includes_min:
                    return other.includes_min
            elif self.min or other.min:
                return bool(other.min)

            if self.max and other.max:
                if self.max != other.max:
                    return self.max < other.max
                return other.includes_max

            return bool(self.max)

        raise NotImplemented()

    def __eq__(self, other):
        return isinstance(other, StandardVersionRange) and self.includes_min == other.includes_min \
               and self.includes_max == other.includes_max and self.min == other.min and self.max == other.max

    def __str__(self):
        res = ''
        if self.min is not None:
            res += '>'
            if self.includes_min:
                res += '='
            res += str(self.min)

        if self.max is not None:
            if res:
                res += ', '
            res += '<'
            if self.includes_max:
                res += '='
            res += str(self.max)

        return res

    def _union_with(self, other: StandardVersionRange) -> VersionSpecifier:
        if other < self:
            return other._union_with(self)

        imin, imax = self.includes_min, other.includes_max
        if not other.min or not self.max or other.min < self.max:
            if self.min == other.min:
                imin = self.includes_min or other.includes_min
            if self.max == other.max:
                imax = self.includes_max or other.includes_max
        elif (not self.includes_max and not other.includes_min) or other.min != self.max:
            return VersionsUnion([self, other])

        return VersionSpecifier.create_range(self.min, other.max, imin, imax)

    def union_with(self, other: VersionSpecifier) -> VersionSpecifier:
        if isinstance(other, (_RestrictAll, _AllowAll, VersionsUnion, HetroVersionIntersection, VersionMatch)):
            return other.union_with(self)

        if isinstance(other, StandardVersionRange):
            if self == other:
                return self
            return self._union_with(other)

        raise NotImplemented()

    def _intersect_with(self, other: StandardVersionRange) -> VersionSpecifier:
        if other < self:
            return other._intersect_with(self)

        if not other.min or not self.max or other.min < self.max:
            imin, imax = other.includes_min, self.includes_max
            if self.min == other.min:
                imin = self.includes_min and other.includes_min
            if self.max == other.max:
                imax = self.includes_max and other.includes_max

            max_ = self.max
            if not max_ or (other.max and self.max > other.max):
                max_, imax = other.max, other.includes_max

            return VersionSpecifier.create_range(other.min, max_, imin, imax)
        if self.includes_max and other.includes_min and self.max == other.min:
            return VersionMatch(self.max, True)

        return RestrictAllVersions

    def intersect_with(self, other: VersionSpecifier) -> VersionSpecifier:
        if isinstance(other, (_RestrictAll, _AllowAll, VersionsUnion, HetroVersionIntersection, VersionMatch)):
            return other.intersect_with(self)

        if isinstance(other, StandardVersionRange):
            if self == other:
                return self

            return self._intersect_with(other)
        raise NotImplemented()

    def allows_version(self, version: Version) -> bool:
        min_, max_ = self.min, self.max
        if (min_ is not None and self.includes_min and min_ == version) \
                or (max_ is not None and self.includes_max and max_ == version):
            return True

        if version.is_post_release() and (min_ is None or not min_.is_post_release()):
            return False

        return (min_ is None or min_ < version) and (max_ is None or version < max_)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Dict, Optional, Union, Set, TYPE_CHECKING

from pkm.api.dependencies.dependency import Dependency
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.packages.site_packages import InstalledPackage
from pkm.api.packages.site_packages import SitePackages
from pkm.api.repositories.repository import Repository, AbstractRepository
from pkm.api.versions.version import StandardVersion
from pkm.api.versions.version_specifiers import VersionMatch
from pkm.resolution.dependency_resolver import resolve_dependencies
from pkm.resolution.pubgrub import UnsolvableProblemException
from pkm.utils.delegations import delegate
from pkm.utils.iterators import first_or_none, single_or_raise
from pkm.utils.promises import Promise
from pkm.utils.properties import cached_property, clear_cached_properties

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment
    from pkm.api.environments.containerized_apps import ContainerizedApplications


@dataclass
class PackageInstallationTarget:  # TODO: maybe rename into package installation site?
    env: "Environment"
    stdlib: str
    platstdlib: str
    platinclude: str
    purelib: str
    platlib: str
    include: str
    data: str
    scripts: str

    @cached_property
    def site_packages(self) -> "SitePackages":
        from pkm.api.packages.site_packages import SitePackages
        return SitePackages(self.env, Path(self.purelib), Path(self.platlib))

    @cached_property
    def app_containers(self) -> "ContainerizedApplications":
        from pkm.api.environments.containerized_apps import ContainerizedApplications
        return ContainerizedApplications(self)

    def reload(self):
        clear_cached_properties(self)

    def uninstall(self, packages_to_remove: List[str]) -> Set[str]:
        """
        attempt to remove the required packages from this target together will all the dependencies
        that may become orphan as a result of this step.

        if a package `p in packages` is a dependency (directly or indirectly) of another
        "user requested" package `q not in packages` then `p` will be kept in the target but its
        "user requested" flag will be removed (if it was existed)

        :param packages_to_remove: the package names to remove
        :return the set of package names that were successfully removed from the environment
        """

        self.reload()

        preinstalled_packages = list(self.site_packages.installed_packages())
        requested_deps = {p.name: p.user_request for p in preinstalled_packages if p.user_request}
        for package_name in packages_to_remove:
            requested_deps.pop(package_name, None)

        user_request = _UserRequestPackage(list(requested_deps.values()))
        installation_repo = _RemovalRepository(preinstalled_packages, user_request)

        installation = resolve_dependencies(user_request.to_dependency(), self.env, installation_repo)
        InstallationPlan(self, installation, {}).execute()

        kept = {p.name for p in installation}

        for p in packages_to_remove:
            if p in kept:
                self.site_packages.installed_package(p).dist_info.unmark_as_user_requested()

        self.reload()
        return {p for p in packages_to_remove if p not in kept}

    def install(
            self, dependencies: List[Dependency], repository: Optional[Repository] = None, user_requested: bool = True,
            dependencies_override: Optional[Dict[str, Dependency]] = None, editables: Optional[Dict[str, bool]] = None,
            updates: Optional[List[str]] = None):
        """
        installs the given set of dependencies into this target.
        see: `prepare_installation` and `PreparedInstallation:install` for more information about this method arguments
        """
        if not dependencies:
            return  # nothing to do...

        repository = repository or self.env.attached_repository
        plan = self.plan_installation(dependencies, repository, user_requested, dependencies_override, updates)
        plan.editables = editables
        plan.execute()

    def force_remove(self, package: str):
        """
        forcefully remove the required package, will not remove its dependencies and will not check if other packages
        depends on it - use this method with care (or don't use it at all :) )
        :param package: the name of the package to be removed
        """
        if installed := self.site_packages.installed_package(package):
            installed.uninstall()

    def plan_installation(
            self, dependencies: List[Dependency], repository: Repository,
            user_requested: bool = True, dependencies_override: Optional[Dict[str, Dependency]] = None,
            updates: Optional[List[str]] = None
    ) -> InstallationPlan:
        """
        plan but does not install an installation for the given dependencies.
        resolve the `dependencies` using the given `repository`, making sure to not break any pre-installed
        "user-requested" packages (but may upgrade their dependencies if it needs to)

        :param dependencies: the dependency to install
        :param repository: the repository to fetch this dependency from, if not given will use the attached repository
        :param user_requested: indicator that the user requested this dependency themselves
            (this will be marked on the installation as per pep376)
        :param dependencies_override: mapping from package name into dependency that should be "forcefully"
            used for this package
        :param updates: If given, the packages listed will be updated if required and already installed
        """

        self.reload()

        updates = set(updates or [])
        preinstalled_packages = [p for p in self.site_packages.installed_packages() if p.name not in updates]
        pre_requested_deps = {p: p.user_request for p in preinstalled_packages if p.user_request}

        new_deps = {d.package_name: d for d in dependencies}
        all_deps = {**pre_requested_deps, **new_deps}

        run_slow_path = False
        installation = None

        try:
            # first we try the fast path: only adding packages without updating
            user_request = _UserRequestPackage(list(new_deps.values()))
            installation_repo = _InstallationRepository(repository, preinstalled_packages, user_request, True)
            installation = resolve_dependencies(
                user_request.to_dependency(), self.env, installation_repo, dependencies_override)

            installation_names = {i.name for i in installation}
            for preinstalled in preinstalled_packages:
                if preinstalled.name not in installation_names:
                    installation.append(preinstalled)

        except UnsolvableProblemException:
            # don't execute the slow path code here so that the stacktrace will be cleaner if exception will happen
            run_slow_path = True

        if run_slow_path:
            # if we cannot we try the slow path in which we allow preinstalled packages dependencies to be updated
            user_request = _UserRequestPackage(list(all_deps.values()))
            installation_repo = _InstallationRepository(repository, preinstalled_packages, user_request, False)
            installation = resolve_dependencies(
                user_request.to_dependency(), self.env, installation_repo)

        return InstallationPlan(self, installation, new_deps if user_requested else {})


class PackageOperation(Enum):
    INSTALL = "install"
    UPDATE = "update"
    REMOVE = "remove"
    SKIP = "skip"


class InstallationPlan:
    def __init__(self, target: PackageInstallationTarget, packages: List[Package],
                 user_requests: Dict[str, "Dependency"]):
        self.default_target = target
        self.packages = packages
        self.user_requests = user_requests
        self.editables: Optional[Dict[str, bool]] = None

    def selected_package(self, name: str) -> Optional[Package]:
        return first_or_none(it for it in self.packages if it.name == name)

    def compute_operations_for_target(
            self, target: Optional[PackageInstallationTarget] = None) -> Dict[Package, PackageOperation]:

        operations: Dict[Package, PackageOperation] = {}

        target = target or self.default_target
        editables = self.editables or {}

        preinstalled: Dict[str, InstalledPackage] = {
            SitePackages.normalize_package_name(p.name): p
            for p in target.site_packages.installed_packages()}

        toinstall: Dict[str, Package] = {
            SitePackages.normalize_package_name(p.name): p
            for p in self.packages
            if not isinstance(p, _UserRequestPackage)}

        editables = {SitePackages.normalize_package_name(name): value for name, value in editables.items()} \
            if editables else {}

        for norm_package_name, package_to_install in toinstall.items():

            if preinstalled_package := preinstalled.pop(norm_package_name, None):
                prev_install_editable = preinstalled_package.installation_info.editable
                editable = editables.get(norm_package_name, prev_install_editable)
                if preinstalled_package.version == package_to_install.version and editable == prev_install_editable:
                    operations[package_to_install] = PackageOperation.SKIP
                    continue

                operations[package_to_install] = PackageOperation.UPDATE
            else:
                operations[package_to_install] = PackageOperation.INSTALL

        for package_to_remove in preinstalled.values():
            operations[package_to_remove] = PackageOperation.REMOVE

        return operations

    def execute(self, target: Optional[PackageInstallationTarget] = None):
        """
        executes the prepared installation inside the given `target`
        :param target: the site in which to execute this installation
        """

        from pkm.api.pkm import pkm

        target = target or self.default_target
        operations = self.compute_operations_for_target(target)
        editables = self.editables or {}
        site = target.site_packages

        promises: List[Promise] = []

        for package, operation in operations.items():
            editable = editables.get(package.name)
            if operation == PackageOperation.INSTALL:
                promises.append(Promise.execute(
                    pkm.threads, package.install_to, target, editable=bool(editable),
                    user_request=self.user_requests.get(package.name)))
            elif operation == PackageOperation.UPDATE:
                if editable is None:
                    editable = site.installed_package(package.name).installation_info.editable
                promises.append(Promise.execute(
                    pkm.threads, package.update_at, target, editable=editable,
                    user_request=self.user_requests.get(package.name)))
            elif operation == PackageOperation.REMOVE:
                promises.append(Promise.execute(pkm.threads, package.uninstall))

        for promise in promises:
            promise.result()

        site.reload()


class _UserRequestPackage(Package):
    def __init__(self, request: List["Dependency"]):
        self._desc = PackageDescriptor("installation request", StandardVersion(release=(0,)))
        self._request = request

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._desc

    def dependencies(
            self, environment: "Environment", extras: Optional[List[str]] = None) -> List["Dependency"]:
        return self._request

    def is_compatible_with(self, env: "Environment") -> bool: return True

    def install_to(self, *args, **kwargs): pass

    def to_dependency(self) -> "Dependency":
        return Dependency(self.name, VersionMatch(self.version))


class _RemovalRepository(AbstractRepository):

    def __init__(self, preinstalled: List[InstalledPackage], user_request: Package):
        super().__init__('removal repository')
        self._preinstalled: Dict[str, InstalledPackage] = {
            PackageDescriptor.normalize_src_package_name(p.name).lower(): p
            for p in preinstalled
        }

        self._user_request = user_request

    def _do_match(self, dependency: "Dependency", env: Environment) -> List[Package]:
        if dependency.package_name == self._user_request.name:
            return [self._user_request]

        return [self._preinstalled[PackageDescriptor.normalize_src_package_name(dependency.package_name).lower()]]


@delegate(Repository, '_repo')
class _InstallationRepository(Repository, ABC):
    def __init__(
            self, repo: Repository, installed_packages: List[InstalledPackage], user_request: Package,
            limit_to_installed: bool):

        assert repo, "no repository provided"
        self._user_request = user_request
        self._installed_packages: Dict[str, InstalledPackage] = {p.name: p for p in installed_packages}
        self._limit_to_installed = limit_to_installed
        self._repo = repo

    def match(self, dependency: Union["Dependency", str], env: Environment) -> List[Package]:
        if isinstance(dependency, str):
            dependency = Dependency.parse(dependency)

        if dependency.package_name == self._user_request.name:
            return [self._user_request]

        if installed := self._installed_packages.get(dependency.package_name):
            installed = _UpdatableInstalledPackage(installed, self._repo)

            if self._limit_to_installed:
                return [installed]

        packages = self._repo.match(dependency, env)
        if installed:
            packages.sort(key=lambda it: 0 if installed.version == it.version else 1)

        return packages


@delegate(Package, '_installed')
class _UpdatableInstalledPackage(Package, ABC):
    def __init__(self, installed: InstalledPackage, repo: Repository):
        self._installed = installed
        self.repo = repo

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._installed.descriptor

    @property
    def published_metadata(self) -> Optional["PackageMetadata"]:
        return self._installed.published_metadata

    def update_at(self, target: "PackageInstallationTarget", user_request: Optional["Dependency"] = None,
                  editable: bool = True):
        new_ver = single_or_raise(self.repo.match(self._installed.descriptor.to_dependency(), target.env))
        user_request = self._installed.user_request if user_request is None else user_request
        self._installed.uninstall()
        new_ver.install_to(target, user_request=user_request, editable=editable)



import compileall
import importlib.util
import re
import warnings
from dataclasses import replace
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, TYPE_CHECKING, List
from zipfile import ZipFile

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distinfo import DistInfo, PackageInstallationInfo, RecordsFileConfiguration
from pkm.api.distributions.distribution import Distribution
from pkm.api.packages.package import PackageDescriptor
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.versions.version import StandardVersion
from pkm.api.versions.version_specifiers import StandardVersionRange
from pkm.distributions.executables import Executables
from pkm.utils.archives import extract_archive
from pkm.utils.files import path_to, CopyTransaction, temp_dir

_METADATA_FILE_RX = re.compile("[^/]*\\.dist-info/METADATA")

if TYPE_CHECKING:
    from pkm.api.packages.package import PackageInstallationTarget
    from pkm.api.projects.project import Project
    from pkm.api.environments.environment import Environment


class InstallationException(IOError):
    ...


class WheelDistribution(Distribution):

    def __init__(self, package: PackageDescriptor, wheel: Path):
        self._wheel = wheel
        assert wheel, "no path for wheel provided"
        self._package = package

    def extract_metadata(self, env: Optional["Environment"] = None) -> PackageMetadata:
        with ZipFile(self._wheel) as zipf:
            for name in zipf.namelist():
                if _METADATA_FILE_RX.fullmatch(name):
                    with TemporaryDirectory() as tdir:
                        zipf.extract(name, tdir)
                        return PackageMetadata.load(Path(tdir) / name)
        raise FileNotFoundError("could not find metadata in wheel")

    @property
    def owner_package(self) -> PackageDescriptor:
        return self._package

    def compute_compatibility_tags(self) -> str:
        """
        return the string that represents the compatibility tags in this wheel file name
        :return: the compatibility tag
        """
        return WheelDistribution.extract_compatibility_tags_of(self._wheel)

    @classmethod
    def extract_compatibility_tags_of(cls, wheel: Path) -> str:
        """
        return the string that represents the compatibility tags in the wheel file name
        :param wheel: the wheel file name
        :return: the compatibility tag
        """
        return '-'.join(wheel.stem.split('-')[-3:])

    @staticmethod
    def expected_wheel_file_name(project: "Project") -> str:
        from pkm.api.environments.environment import Environment
        project_config = project.config.project
        req = project_config.requires_python

        min_interpreter: StandardVersion = req.min \
            if req and isinstance(req, StandardVersionRange) else \
            StandardVersion((Environment.current().interpreter_version.release[0],))

        req_interpreter = 'py' + ''.join(str(it) for it in min_interpreter.release[:2])
        return f"{project.descriptor.expected_src_package_name}-{project.version}-{req_interpreter}-none-any.whl"

    @classmethod
    def install_extracted_wheel(
            cls, package: PackageDescriptor, content: Path, target: "PackageInstallationTarget",
            user_request: Optional[Dependency] = None, installation_info: Optional[PackageInstallationInfo] = None,
            skip_record_verification: bool = False):

        dist_info = _find_dist_info(content, package)

        wheel_file = dist_info.load_wheel_cfg()
        wheel_file.validate_supported_version()

        entrypoints = dist_info.load_entrypoints_cfg().entrypoints

        site_packages = Path(target.purelib if wheel_file['Root-Is-Purelib'] == 'true' else target.platlib)

        records_file: RecordsFileConfiguration = dist_info.load_record_cfg()
        if not skip_record_verification:
            _verify_records(dist_info, content, records_file)

        with CopyTransaction() as ct:
            for d in content.iterdir():
                if d.is_dir():
                    if d.suffix == '.data':
                        for k in d.iterdir():
                            if not (target_path := getattr(target, k.name, None)):
                                raise InstallationException(
                                    f'wheel contains data entry with unsupported key: {k.name}')

                            if k.name == 'scripts':
                                ct.copy_tree(
                                    k, Path(target_path),
                                    file_copy=lambda s, t: Executables.patch_shabang_for_interpreter(
                                        s, t, target.env.interpreter_path))
                            else:
                                ct.copy_tree(k, Path(target_path))
                    else:
                        ct.copy_tree(d, site_packages / d.name, accept=lambda it: it != records_file.path)
                else:
                    ct.copy(d, site_packages / d.name)

            # build entry points
            scripts_path = Path(target.scripts)
            for entrypoint in entrypoints:
                if entrypoint.is_script():
                    ct.touch(Executables.generate_for_entrypoint(target.env, entrypoint, scripts_path))

            #  compile py to pyc
            new_pyc_files: List[str] = []
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore')
                for cc in ct.copied_files:
                    if cc.suffix == '.py':
                        compileall.compile_file(cc, force=True, quiet=2)
                        new_pyc_files.append(importlib.util.cache_from_source(str(cc)))

            # mark the compiled files in the transaction, so that they will be added to record
            for npf in new_pyc_files:
                ct.touch(Path(npf))

            # build the new records file
            new_dist_info = DistInfo.load(site_packages / dist_info.path.name)
            new_record_file = new_dist_info.load_record_cfg()

            precomputed_hashes = {
                r.file: r.hash_signature for r in records_file.records
            } if not skip_record_verification else None
            new_record_file.sign_files(ct.copied_files, site_packages, precomputed_hashes)

            new_record_file.save()

            # and finally, mark the installer and the requested flag
            (new_dist_info.path / "INSTALLER").write_text(f"pkm")
            if user_request:
                new_dist_info.mark_as_user_requested(user_request)
            if installation_info:
                new_dist_info.save_installation_info(installation_info)

    def install_to(self, target: "PackageInstallationTarget", user_request: Optional[Dependency] = None,
                   installation_mode: Optional[PackageInstallationInfo] = None):
        """
        Implementation of wheel installer based on PEP427
        as described in: https://packaging.python.org/en/latest/specifications/binary-distribution-format/
        """
        with temp_dir() as tmp_path:
            extract_archive(self._wheel, tmp_path)
            if not installation_mode:
                installation_mode = PackageInstallationInfo()

            if not installation_mode.compatibility_tag:
                installation_mode = replace(
                    installation_mode, compatibility_tag=WheelDistribution.extract_compatibility_tags_of(self._wheel))
                WheelDistribution.install_extracted_wheel(self._package, tmp_path, target, user_request,
                                                          installation_mode)


def _find_dist_info(unpacked_wheel: Path, package: PackageDescriptor) -> DistInfo:
    dist_info = list(unpacked_wheel.glob("*.dist-info"))
    if not dist_info:
        raise InstallationException(f"wheel for {package} does not contain dist-info")
    if len(dist_info) != 1:
        raise InstallationException(f"wheel for {package} contains more than one possible dist-info")

    return DistInfo.load(dist_info[0])


def _verify_records(dist_info: DistInfo, content: Path, records_file: RecordsFileConfiguration):
    if not records_file.exists():
        raise InstallationException(
            f"Unsigned wheel for package {dist_info.package_name} (no RECORD file found in dist-info)")

    # check that the records hash match
    record_by_path = {r.file: r for r in records_file.records}
    for file in content.rglob("*"):
        if file.is_dir():
            continue

        path = str(path_to(content, file))
        if record := record_by_path.get(path):
            if not record.hash_signature.validate_against(file):
                if any(it.name.endswith('.dist-info') for it in file.parents):
                    warnings.warn(f"mismatch hash signature for {file}")
                else:
                    raise InstallationException(f"File signature not matched for: {record.file}")

        elif file != dist_info.path / "RECORD":
            raise InstallationException(
                f"Wheel contains files with no signature in RECORD, "
                f"e.g., {path}")



from __future__ import annotations

import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory, mkdtemp
from typing import Optional, Callable, ContextManager, Iterator, List, Set

from pkm.utils.commons import UnsupportedOperationException


def is_empty_directory(path: Path) -> bool:
    """
    check if the given `path` is a directory and is empty
    :param path: the path to check
    :return: if `path` is empty directory
    """

    return path.is_dir() and next(path.iterdir(), None) is None


def ensure_exists(path: Path, error_msg: Optional[Callable[[], str]] = None) -> Path:
    """
    :param path: the path to check if exists
    :param error_msg: the error message to use in case that the path does not exist
    :return: the given `path` if it exists, otherwise raise `FileNotFoundError`
    """
    if path.exists():
        return path
    raise FileNotFoundError(f"no such file or directory: {path}" if error_msg is None else error_msg())


def mkdir(path: Path) -> Path:
    """
    make the given path as a directory (including parents), will do nothing if already exists
    :param path: the path to make as a directory
    :return: the given path (allows usage with chaining)
    """
    path.mkdir(parents=True, exist_ok=True)
    return path


def resolve_relativity(path: Path, parent: Path) -> Path:
    """
    :param path: the path to resolve
    :param parent: the potential parent of the path (if the path is relative)
    :return: if `path` is absolute, return `path` otherwise return `parent / path`
    """

    return path if path.is_absolute() else parent / path


def path_to(source: Path, destination: Path) -> Path:
    """
    creates a relative path from `source` to `destination`, allowing back stepping ('..')
    :param source: the path to start from
    :param destination: the path to end at
    :return: the relative path from `source` to `destination`
    """
    source = source.absolute()
    destination = destination.absolute()

    destination_parents = set(destination.parents)
    p = source
    back = 0
    while p not in destination_parents:
        p = p.parent
        back += 1
    return Path((f'..{os.sep}' * back) + str(destination.relative_to(p)))


def is_root_path(path: Path) -> bool:
    """
    :param path: the path to check
    :return: True if the given `path` is a file-system root, False otherwise.
    """
    return path == path.parent


@contextmanager
def temp_dir() -> ContextManager[Path]:
    """
    creates a temporary directory and return a context manager, which, when closing, deletes it
    :return:
    """

    with TemporaryDirectory() as tdir:
        yield Path(tdir)


def extension_of(path: Path) -> str:
    """
    extract the given path extension, unlike `path.suffix` this will capture all composed extensions
    like `.tar.gz`
    :param path: a path to a file, its extension you want to get
    :return: the file extension
    """

    name = path.name
    try:
        return name[name.index('.'):]
    except ValueError:
        return ""


def name_without_ext(path: Path) -> str:
    """
    extract the given path name without extensions, unlike `path.stem` this will remove all composed extensions
    like `.tar.gz`
    :param path: a path to a file, its name you want to get
    :return: the file name without any extension (read: until the first '.')
    """

    name = path.name
    try:
        return name[:name.index('.')]
    except ValueError:
        return name


class CopyTransaction:
    """
    a context manager for copy transactions,
    this keep track on a transaction of copy operations which may also overwrite existing files,
    if an uncaught exception was raised during this transaction execution, the operations are rollback
    and any overwritten file is restored
    """

    def __init__(self):
        self._copied_files: Set[Path] = set()
        self._overwritten_files: List[Path] = []  # source overwrite, temp location
        self._temp_dir: Optional[Path] = None

    def copy_tree(self, source: Path, target: Path,
                  accept: Callable[[Path], bool] = lambda _: True,
                  file_copy: Optional[Callable[[Path, Path], None]] = None):
        """
        copy the given source directory files into the target directory (creating it if necessary)
        :param source: a directory to copy files from
        :param target: a directory to copy files to
        :param accept: a predicate that is called before each copy operation, if it returns False,
                       the copy will not be made
        :param file_copy: the function that will be used to copy a file
        """

        if file_copy:
            def cp(s, d):
                file_copy(s, d)
                self.touch(d)
        else:
            cp = self.copy

        self.mkdir(target)
        for file in source.iterdir():
            if not accept(file):
                continue

            if file.is_dir():
                self.copy_tree(file, target / file.name, accept)
            else:
                cp(file, target / file.name)

    def mkdir(self, target: Path) -> Path:
        """
        creates a directory in the given `target` path
        :param target: the directory path to create
        :return `target` to be used with chaining
        """
        if not target.exists():
            if not target.parent.exists():
                self.mkdir(target.parent)
            target.mkdir()
            self._copied_files.add(target)

        return target

    def copy(self, source: Path, target: Path):
        """
        copies the `source` file into the `target` path, if the `target` already exists, overwrite it
        :param source: the source file to copy
        :param target: the target path to copy into
        """
        if source.is_dir():
            raise UnsupportedOperationException("copy should not work with directories, use copy_tree or mkdir instead")

        if target.exists():
            self.rm(target)

        try:
            shutil.copy(source, target)
        except FileNotFoundError:
            if not target.exists():
                self.mkdir(target.parent)
                shutil.copy(source, target)
            else:
                raise

        self._copied_files.add(target)

    @property
    def copied_files(self) -> Iterator[Path]:
        """
        :return: list of all the files (including directories) that was created as a result of this transaction
        """
        return iter(self._copied_files)

    def touch_all(self, paths: List[Path]):
        """
        mark the given `paths` as copied in this transaction
        :param paths: list of path to mark
        """
        self._copied_files.update(paths)

    def touch(self, path: Path, fs_touch: bool = False):
        """
        mark the given `path` as copied in this transaction
        :param path: the path to mark
        :param fs_touch: if true will also touch the path in the filesystem
        """
        self._copied_files.add(path)
        if fs_touch:
            path.touch(exist_ok=True)

    def touch_tree(self, path: Path):
        """
        mark the given `path` and all its children as copied in this transaction
        :param path: the path to mark
        """
        self.touch(path)
        self._copied_files.update(path.rglob("*"))

    def rm(self, path: Path):
        """
        remove the given path (rollback supported operation)
        works for both files and directories
        :param path: the path to remove
        """
        temp = self._temp_dir / str(len(self._overwritten_files))
        shutil.move(path, temp)
        self._overwritten_files.append(path)

    def __enter__(self) -> CopyTransaction:
        self._temp_dir = Path(mkdtemp())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:  # rollback
            import traceback
            traceback.print_exc()
            self._rollback()
        else:
            self._commit()

    def _commit(self):
        shutil.rmtree(self._temp_dir)
        self._temp_dir = None

    def _rollback(self):
        for file in self._copied_files:
            if not file.is_dir():
                file.unlink(missing_ok=True)

        for file in self._copied_files:
            if file.is_dir() and is_empty_directory(file):
                file.rmdir()

        for i, path in enumerate(self._overwritten_files):
            restore_path = self._temp_dir / str(i)
            shutil.move(restore_path, path)

        self._commit()


def is_relative_to(path: Path, root: Path) -> bool:
    """
    tests if `path` is relative (read "resides in") root
    :param path: the path to check
    :param root: the path that may contain `path`
    """

    return str(path.absolute()).startswith(str(root.absolute()))



from pathlib import Path
from typing import List, Iterable

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.standard_package import PackageArtifact, AbstractPackage
from pkm.api.pkm import pkm
from pkm.api.repositories.repository import AbstractRepository
from pkm.utils.http.http_client import Url


class UrlRepository(AbstractRepository):

    def __init__(self):
        super().__init__('url')

    def accepted_url_protocols(self) -> Iterable[str]:
        return 'url', 'http', 'https'

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        if not (vurl := dependency.required_url()):
            return []

        return [UrlPackage(PackageDescriptor(dependency.package_name, vurl), vurl.url)]


class UrlPackage(AbstractPackage):
    def __init__(self, desc: PackageDescriptor, url: str):
        purl = Url.parse(url)
        file_name = Path(purl.path).name
        super().__init__(desc, [PackageArtifact(file_name)])
        self._url = url

    def _retrieve_artifact(self, artifact: PackageArtifact) -> Path:
        return pkm.httpclient.fetch_resource(str(self._url)).data



# from typing import Iterable, TYPE_CHECKING
#
# from pkm.api.repositories.repository import Repository
# from pkm.utils.monitors import no_monitor, Monitor
#
# if TYPE_CHECKING:
#     from pkm.api.dependencies.dependency import Dependency
#     from pkm.api.projects.project import Project
#     from pkm.api.projects.project_group import ProjectGroup
#
#
# # noinspection PyMethodMayBeStatic,PyUnusedLocal
# class ProjectOperationsMonitor():
#     def on_environment_modification(self, project: "Project") -> EnvironmentOperationsMonitor:
#         return no_monitor()
#
#     def on_dependencies_modified(self, project: "Project", old_deps: Iterable["Dependency"],
#                                  new_deps: Iterable["Dependency"]):
#         ...
#
#     def on_lock_modified(self):
#         ...
#
#     def on_pyproject_modified(self):
#         ...
#
#     def on_publish(self, repo: Repository) -> Monitor:
#         return no_monitor()
#
#
# # noinspection PyMethodMayBeStatic,PyUnusedLocal
# class ProjectGroupOperationsMonitor(Monitor):
#     def on_access_child_project(self, group: "ProjectGroup", child_project: "Project") -> ProjectOperationsMonitor:
#         return no_monitor()
#
#     def on_access_child_group(
#             self, group: "ProjectGroup", child_group: "ProjectGroup") -> "ProjectGroupOperationsMonitor":
#         return no_monitor()
#
#     def on_build_all(self, group: "ProjectGroup") -> Monitor:
#         return no_monitor()
#
#     def on_install_all(self, group: "ProjectGroup") -> Monitor:
#         return no_monitor()
#
#     def on_publish_all(self, group: "ProjectGroup") -> Monitor:
#         return no_monitor()



import os
import re
import shutil
from pathlib import Path
from typing import Dict, Union

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.environments.environment_introspection import EnvironmentIntrospection
from pkm.api.pkm import pkm
from pkm.config.configuration import FileConfiguration
import sys

from pkm.utils.commons import NoSuchElementException

_INTERPRETER_INTROSPECTIONS: Dict[str, EnvironmentIntrospection] = {}
_PYVENV_SEP_RX = re.compile("\\s*=\\s*")


class EnvironmentBuilder:

    @staticmethod
    def create_matching(env_path: Path, interpreter: Dependency) -> Environment:
        result = Environment(env_path)

        python_versions = pkm.repositories.installed_pythons.match(interpreter, result)
        if not python_versions:
            raise NoSuchElementException("could not find installed python interpreter "
                                         f"that match the given dependency: {interpreter}")
        python_versions[0].install_to(result)
        return result

    @staticmethod
    def create(env_path: Path, interpreter_path: Path = Path(sys.executable)) -> Environment:
        interpreter_path = interpreter_path.absolute()

        if env_path.exists():
            raise FileExistsError(f"{env_path} already exists")

        ispc = _introspection_for(interpreter_path)
        sys_platform = ispc['sys.platform']
        is_windows = ispc.is_windows_env()
        sys_vinfo = ispc['sys.version_info']

        env_path.mkdir(parents=True, exist_ok=True)

        # prepare pyvenv.cfg
        pyvenvcfg = PyVEnvConfiguration.load(env_path / 'pyvenv.cfg')
        pyvenvcfg['home'] = str(interpreter_path.parent)
        pyvenvcfg['version'] = '.'.join(str(it) for it in sys_vinfo[:3])
        pyvenvcfg['include-system-site-packages'] = 'false'
        pyvenvcfg.save()

        # build relevant directories
        if is_windows:
            bin_path = env_path / 'Scripts'
            include_path = env_path / 'Include'
            site_packages_path = env_path / 'Lib' / 'site-packages'
        else:
            bin_path = env_path / 'bin'
            include_path = env_path / 'include'
            site_packages_path = env_path / 'lib' / f'python{sys_vinfo[0]}.{sys_vinfo[1]}' / 'site-packages'

        for path in (bin_path, include_path, site_packages_path):
            path.mkdir(exist_ok=True, parents=True)

        if not ispc.is_32bit_interpreter and ispc['os.name'] == 'posix' and sys_platform != 'darwin':
            (env_path / 'lib64').symlink_to(env_path / 'lib')

        # copy needed files
        if is_windows:
            file_names_to_copy = ['python.exe', 'python_d.exe', 'pythonw.exe', 'pythonw_d.exe']
            python_dir = interpreter_path.parent
            for file_name in file_names_to_copy:
                if (python_file := python_dir / file_name).exists():
                    shutil.copy(python_file, bin_path / file_name)

            if ispc['sysconfig.is_python_build']:
                # copy init.tcl
                if init_tcl := next(python_dir.rglob("**/init.tcl"), None):
                    target_init_tcl = env_path / str(init_tcl.relative_to(python_dir))
                    target_init_tcl.parent.mkdir(exist_ok=True, parents=True)
                    shutil.copy(init_tcl, target_init_tcl)
        else:
            for i in range(3):
                executable_name = f"python{'.'.join(str(it) for it in sys_vinfo[:i])}"
                executable_path = bin_path / executable_name
                if not executable_path.exists():
                    executable_path.symlink_to(interpreter_path)

        return Environment(env_path)


def _introspection_for(interpreter_path: Path) -> EnvironmentIntrospection:
    interpreter_key = str(interpreter_path.absolute())
    if (intro := _INTERPRETER_INTROSPECTIONS.get(interpreter_key)) is None:
        _INTERPRETER_INTROSPECTIONS[interpreter_key] = intro = EnvironmentIntrospection.compute(interpreter_path)

    return intro


class PyVEnvConfiguration(FileConfiguration):

    def generate_content(self) -> str:
        return os.linesep.join(f"{k} = {v}" for k, v in self._data.items())

    @classmethod
    def load(cls, path: Union[Path, str]):
        path = Path(path)

        if not path.exists():
            return PyVEnvConfiguration(path=path, data={})

        data = {(kv := _PYVENV_SEP_RX.split(line))[0]: kv[1]
                for line in path.read_text().splitlines(keepends=False)}
        return PyVEnvConfiguration(path=path, data=data)
